@@ -4,6 +4,8 @@ from ._datasets import (
     blobs,
     wiki_sp500_members_update,
     download_tickers_price_history,
+    download_tickers_price_history_fromlist,
     get_wiki_sp500_metadata,
     download_tickers_fundamental,
+    process_eod_price,
 )
